#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <getopt.h>

#include "ProductVer.h"
#include "audio-device-info.h"
#include "device-property.h"

typedef enum _MWCAP_INFO_CMD_
{
    MWCAP_INFO_USAGE,
    MWCAP_INFO_LIST_CHANNEL_INFO,

    MWCAP_INFO_QUERY,

    MWCAP_INFO_ALL,
    MWCAP_INFO_DEVICE,
    MWCAP_INFO_INPUT_ALL,
    MWCAP_INFO_INPUT_VIDEO,
    MWCAP_INFO_INPUT_AUDIO,
    MWCAP_INFO_INPUT_SPECIFIC,
    MWCAP_INFO_HDMI,

    MWCAP_INFO_V4L2_NAME,
    MWCAP_INFO_ALSA_NAME,
    MWCAP_INFO_ALSA_NAME_MPLAYER,
    MWCAP_INFO_ALSA_NAME_PA,


    MWCAP_INFO_DEVICE_NUMS,
}MWCAP_INFO_CMD;

static DWORD g_dwCMD = 0;

static int g_nBoardIndex = -1;
static int g_nChannelIndex = -1;
static char g_arrDevicesPath[MAX_CHANNEL_COUNT][MAX_CHANNEL_NAME_LEN];
static int g_nDevicesNum = 0;


static const char short_options[] = "lhq:i:v:a:c:";

static const struct option long_options[] = {
    { "info-all",               required_argument, NULL, 'i' },
    { "info-device",            required_argument, NULL,  0  },
    { "info-input-all",         required_argument, NULL,  0  },
    { "info-input-video",       required_argument, NULL,  0  },
    { "info-input-audio",       required_argument, NULL,  0  },
    { "info-input-specific",    required_argument, NULL,  0  },
    { "info-hdmi",              required_argument, NULL,  0  },

    { "alsa-name",              required_argument, NULL, 'a' },
    { "alsa-name-mplayer",      required_argument, NULL,  0  },
    { "alsa-name-pa",           required_argument, NULL,  0  },


    { 0, 0, 0, 0 }
};


void PrintUsage()
{
    printf("USAGE\n"
           "  mwcap-info [OPTIONS]\n\n"
           "DESCRIPTION\n"
           "  Show device, video, audio information of Pro Capture cards.\n\n"
           "OPTIONS\n"
           "  -h                                            print usage (Default option)\n"
           "  -l                                            list all capture device information\n\n"
           "  -q CAPTURE-DEVICE                             query capture device information, CAPTURE-DEVICE can be\n"
           "                                                V4L2 device path (eg. /dev/video0) or capture device ID\n"
           "                                                (Board ID:Channel ID, eg. 0:0)\n\n"
           "  -i, --info-all CAPTURE-DEVICE                 show all information of Pro Capture cards\n"
           "      --info-device CAPTURE-DEVICE              show device information of Pro Capture cards\n\n"
           "      --info-input-all CAPTURE-DEVICE           show all input information\n"
           "      --info-input-video CAPTURE-DEVICE         show video input information of Pro Capture cards\n"
           "      --info-input-audio CAPTURE-DEVICE         show audio input information of Pro Capture cards\n"
           "      --info-input-specific CAPTURE-DEVICE      show specific input information of Pro Capture cards\n\n"
           "  -v DEVICE-ID                                  show V4L2 device path of specified capture device ID\n"
           "  -a, --alsa-name CAPTURE-DEVICE                show ALSA name of specified capture device\n"
           "      --alsa-name-mplayer CAPTURE-DEVICE        show ALSA name of specified capture device in mplayer format\n"
           "      --alsa-name-pa CAPTURE-DEVICE             show ALSA name of specified capture device in paulse audio\n"
           "                                                format\n"
           "  -c DEVICE-ID                                  show devices' counts of specified capture device ID\n\n"
           "EXAMPLE\n"
           "  1. list all capture device information\n"
           "       $ mwcap-info -l\n"
           "  2. query capture device information\n"
           "       $ mwcap-info -q /dev/video0\n"
           "       -- or --\n"
           "       $ mwcap-info -q 0:0\n"
           "  3. show all information of Pro Capture cards\n"
           "       $ mwcap-info -i /dev/video0\n"
           "       -- or --\n"
           "       $ mwcap-info -i 0:0\n"
           "       -- or --\n"
           "       $ mwcap-info --info-all /dev/video0\n"
           "       -- or --\n"
           "       $ mwcap-info --info-all 0:0\n"
           "  4. show device information of Pro Capture cards\n"
           "       $ mwcap-info --info-device /dev/video0\n"
           "       -- or --\n"
           "       $ mwcap-info --info-device 0:0\n"
           "  5. show all input information of Pro Capture cards\n"
           "       $ mwcap-info --info-input-all /dev/video0\n"
           "       -- or --\n"
           "       $ mwcap-info --info-input-all 0:0\n"
           "  6. show video input information of Pro Capture cards\n"
           "       $ mwcap-info --info-input-video /dev/video0\n"
           "       -- or --\n"
           "       $ mwcap-info --info-input-video 0:0\n"
           "  7. show audio input information of Pro Capture cards\n"
           "       $ mwcap-info --info-input-audio /dev/video0\n"
           "       -- or --\n"
           "       $ mwcap-info --info-input-audio 0:0\n"
           "  8. show specific input information of Pro Capture cards\n"
           "       $ mwcap-info --info-input-specific /dev/video0\n"
           "       -- or --\n"
           "       $ mwcap-info --info-input-specific 0:0\n"
           "  9. show V4L2 device path of capture device0\n"
           "       $ mwcap-info -v 0:0\n"
           "  10. show ALSA name of specified capture device0\n"
           "       $ mwcap-info -a /dev/video\n"
           "       -- or --\n"
           "       $ mwcap-info -a 0:0\n"
           "       -- or --\n"
           "       $ mwcap-info --alsa-name /dev/video0\n"
           "       -- or --\n"
           "       $ mwcap-info --alsa-name 0:0\n"
           "  11. show ALSA name of specified capture device in mplayer format\n"
           "       $ mwcap-info --alsa-name-mplayer /dev/video0\n"
           "       -- or --\n"
           "       $ mwcap-info --alsa-name-mplayer 0:0\n"
           "  12. show ALSA name of specified capture device in paulse audio format\n"
           "       $ mwcap-info --alsa-name-pa /dev/video0\n"
           "       -- or --\n"
           "       $ mwcap-info --alsa-name-pa 0:0\n"
           "  13. show devices counts of specified capture device ID \n"
           "       $ mwcap-info -c 0:0\n"
           );
}

static bool IsProCaptureDevice(char *szDevicePath)
{
    if (szDevicePath == NULL)
        return false;
    int nCount = MWGetChannelCount();

    for (int i = 0; i < nCount; i++)
    {
        char szTemp[MAX_CHANNEL_NAME_LEN];
        MWGetDevicePath(i, szTemp);

        if (strcmp(szTemp, szDevicePath) == 0)
            return true;
    }

    return false;
}

static bool IsAlphaNumeric(char *str)
{
    if (str == NULL)
        return false;

    char *temp = str;

    while(*temp)
    {
        if ( *str >= '0' && *str <= '9')
            return true;

        temp++;
    }

    return false;
}

static MWCAP_INFO_RESULT PraiseDeviceArg(char *szArg, char (*arrDevicesPath)[MAX_CHANNEL_NAME_LEN], int *nDevicesNum)
{
    MWCAP_INFO_RESULT ir;

    if (szArg == NULL)
        return RESULT_INVALID_PARAMS;

    int nTemp = 0;

    if (strchr(szArg, ':') == NULL)
    {
        if (access(szArg, F_OK) != 0)
            return RESULT_NOT_EXIST_DEVICE;

        if (!IsProCaptureDevice(szArg))
            return RESULT_NOT_PRO_CAPTURE_DEVICE;

        if (nTemp < *nDevicesNum)
            strcpy(arrDevicesPath[nTemp], szArg);

        nTemp++;
        *nDevicesNum = nTemp;
    }
    else
    {
        char *p;
        char *temp = strdup(szArg);
        char *separator = ":";
        int nCount = 0;

        p = strtok(temp, separator);
        while(p != NULL)
        {
            if (!IsAlphaNumeric(p))
                return RESULT_ERROR_ARGUMENT;

            switch(nCount)
            {
            case 0:
            {
                g_nBoardIndex = (BYTE)atoi(p);
                break;
            }
            case 1:
            {
                g_nChannelIndex = (BYTE)atoi(p);
                break;
            }
            default:
                break;
            }
            nCount++;

            p = strtok(NULL, separator);
        }

        if (nCount != 2)          
            return RESULT_ERROR_ARGUMENT;


        free(temp);
        temp = NULL;


        ir = GetDevicePath(arrDevicesPath, nDevicesNum, g_nBoardIndex, g_nChannelIndex);
        if (ir != RESULT_OK)
            return ir;
    }

    return RESULT_OK;
}


static MWCAP_INFO_RESULT PraseCommandLine (int argc, char *argv[])
{
    MWCAP_INFO_RESULT ir;
    for(;;)
    {
        int nIdx;
        int c;
        int nTemp = 64;

        c = getopt_long(argc, argv, short_options, long_options, &nIdx);
        if (c == -1)
            break;
        else if (c == '?')
            return RESULT_ERROR_OPTION;

        switch (c) {
        case 0:
        {
            nTemp = 64;
            if (strcmp(long_options[nIdx].name, "info-all") == 0)
            {
                g_dwCMD |= (0x01 << MWCAP_INFO_ALL);
                ir = PraiseDeviceArg(optarg, g_arrDevicesPath, &nTemp);
                if (ir != RESULT_OK)
                    return ir;

                g_nDevicesNum = nTemp;
            }
            else if (strcmp(long_options[nIdx].name, "info-device") == 0)
            {
                g_dwCMD |= (0x01 << MWCAP_INFO_DEVICE);
                ir = PraiseDeviceArg(optarg, g_arrDevicesPath, &nTemp);
                if (ir != RESULT_OK)
                    return ir;

                g_nDevicesNum = nTemp;
            }
            else if (strcmp(long_options[nIdx].name, "info-input-all") == 0)
            {
                g_dwCMD |= (0x01 << MWCAP_INFO_INPUT_ALL);
                ir = PraiseDeviceArg(optarg, g_arrDevicesPath, &nTemp);
                if (ir != RESULT_OK)
                    return ir;

                g_nDevicesNum = nTemp;
            }
            else if (strcmp(long_options[nIdx].name, "info-input-video") == 0)
            {
                g_dwCMD |= (0x01 << MWCAP_INFO_INPUT_VIDEO);
                ir = PraiseDeviceArg(optarg, g_arrDevicesPath, &nTemp);
                if (ir != RESULT_OK)
                    return ir;

                g_nDevicesNum = nTemp;
            }
            else if (strcmp(long_options[nIdx].name, "info-input-audio") == 0)
            {
                g_dwCMD |= (0x01 << MWCAP_INFO_INPUT_AUDIO);
                ir = PraiseDeviceArg(optarg, g_arrDevicesPath, &nTemp);
                if (ir != RESULT_OK)
                    return ir;

                g_nDevicesNum = nTemp;
            }
            else if (strcmp(long_options[nIdx].name, "info-input-specific") == 0)
            {
                g_dwCMD |= (0x01 << MWCAP_INFO_INPUT_SPECIFIC);
                ir = PraiseDeviceArg(optarg, g_arrDevicesPath, &nTemp);
                if (ir != RESULT_OK)
                    return ir;

                g_nDevicesNum = nTemp;
            }
            else if (strcmp(long_options[nIdx].name, "info-hdmi") == 0)
            {
                g_dwCMD |= (0x01 << MWCAP_INFO_HDMI);
                ir = PraiseDeviceArg(optarg, g_arrDevicesPath, &nTemp);
                if (ir != RESULT_OK)
                    return ir;

                g_nDevicesNum = nTemp;
            }
            else if (strcmp(long_options[nIdx].name, "alsa-name") == 0)
            {
                g_dwCMD |= (0x01 << MWCAP_INFO_ALSA_NAME);
                ir = PraiseDeviceArg(optarg, g_arrDevicesPath, &nTemp);
                if (ir != RESULT_OK)
                    return ir;

                g_nDevicesNum = nTemp;
            }
            else if (strcmp(long_options[nIdx].name, "alsa-name-mplayer") == 0)
            {
                g_dwCMD |= (0x01 << MWCAP_INFO_ALSA_NAME_MPLAYER);
                ir = PraiseDeviceArg(optarg, g_arrDevicesPath, &nTemp);
                if (ir != RESULT_OK)
                    return ir;

                g_nDevicesNum = nTemp;
            }
            else if (strcmp(long_options[nIdx].name, "alsa-name-pa") == 0)
            {
                g_dwCMD |= (0x01 << MWCAP_INFO_ALSA_NAME_PA);
                ir = PraiseDeviceArg(optarg, g_arrDevicesPath, &nTemp);
                if (ir != RESULT_OK)
                    return ir;

                g_nDevicesNum = nTemp;
            }

        }
            break;
        case 'l':
        {
            g_dwCMD |=  (0x01 << MWCAP_INFO_LIST_CHANNEL_INFO);
        }
            break;
        case 'h':
        {
            g_dwCMD |=  (0x01 <<  MWCAP_INFO_USAGE);
        }
            break;
        case 'q':
        {
            nTemp = 64;
            g_dwCMD |= (0x01 << MWCAP_INFO_QUERY);
            ir = PraiseDeviceArg(optarg, g_arrDevicesPath, &nTemp);
            if (ir != RESULT_OK)
                return ir;

            g_nDevicesNum = nTemp;
        }
            break;
        case 'v':
        {
            nTemp = 64;
            g_dwCMD |= (0x01 << MWCAP_INFO_V4L2_NAME);
            ir = PraiseDeviceArg(optarg, g_arrDevicesPath, &nTemp);
            if (ir != RESULT_OK)
                return ir;

            g_nDevicesNum = nTemp;
        }
            break;
        case 'i':
        {
            nTemp = 64;
            g_dwCMD |= (0x01 << MWCAP_INFO_ALL);
            ir = PraiseDeviceArg(optarg, g_arrDevicesPath, &nTemp);
            if (ir != RESULT_OK)
                return ir;

            g_nDevicesNum = nTemp;
        }
            break;
        case 'c':
        {
            nTemp = 64;
            g_dwCMD |= (0x01 << MWCAP_INFO_DEVICE_NUMS);
            ir = PraiseDeviceArg(optarg, g_arrDevicesPath, &nTemp);
            if (ir != RESULT_OK)
                return ir;

            g_nDevicesNum = nTemp;
        }
            break;
        case 'a':
        {
            g_dwCMD |= (0x01 << MWCAP_INFO_ALSA_NAME);
            ir = PraiseDeviceArg(optarg, g_arrDevicesPath, &nTemp);
            if (ir != RESULT_OK)
                return ir;

            g_nDevicesNum = nTemp;
        }
            break;
        default:
            break;
        }
    }

    /*int j;
    for (j = 0; j < g_nDevicesNum; j++)
    {
        if (access(g_arrDevicesPath[j], F_OK) != 0)
        {
            printf("erroe device %s doesn't exit\n", g_arrDevicesPath[j]);
            return false;
        }
    }*/

    return RESULT_OK;
}


int main(int argc, char *argv[])
{
    MWCaptureInitInstance();
    MWRefreshDevice();
    RefreshAudioDevices();



    MW_RESULT mr;
    MWCAP_INFO_RESULT ir;

    //BOOL bRet;
    do{
        ir = PraseCommandLine (argc, argv);

        if (ir != RESULT_OK)
        {
            if ((g_dwCMD & 0x01 << MWCAP_INFO_V4L2_NAME) ||
                    (g_dwCMD & 0x01 << MWCAP_INFO_ALSA_NAME) ||
                    (g_dwCMD & 0x01 << MWCAP_INFO_ALSA_NAME_MPLAYER) ||
                    (g_dwCMD & 0x01 << MWCAP_INFO_ALSA_NAME_PA))
                return -1;
        }

        switch (ir)
        {
        case RESULT_ERROR_ARGUMENT:
        {
            printf("Error argument \"%s\", please use \"-h\" option to see usage\n", argv[optind-1]);
            MWCaptureExitInstance();
            //PrintUsage();
            return -1;
        }
            break;
        case RESULT_ERROR_OPTION:
        {
            printf("Error option %s, please use \"-h\" option to see usage\n", argv[optind - 1]);
            PrintUsage();
            MWCaptureExitInstance();
            return -1;
        }
            break;
        case RESULT_NOT_EXIST_DEVICE:
        {
            printf("Device %s doesn't exist\n", optarg);
            MWCaptureExitInstance();
            return -1;
        }
            break;
        case RESULT_NOT_PRO_CAPTURE_DEVICE:
        {
            printf("Device %s isn't a pro capture device\n", optarg);
            MWCaptureExitInstance();
            return -1;
        }
            break;
        case RESULT_UNKNOW_ERROR:
        {
            MWCaptureExitInstance();
            return -1;
        }
            break;
        case RESULT_NO_MATCHING_DEVICE:
        {
            if ((g_dwCMD & (0x01 << MWCAP_INFO_DEVICE_NUMS)))
                 break;
            else if((g_dwCMD & (0x01 << MWCAP_INFO_QUERY)))
            {
                break;
            }
            else
            {
                printf("No device with board-id: %d, channel-id: %d\n", g_nBoardIndex, g_nChannelIndex);
                MWCaptureExitInstance();
                return -1;
            }
        }
            break;
        case RESULT_OK:
            break;
        default:
            break;
        }

        if ((g_dwCMD & (0x01 << MWCAP_INFO_USAGE)) || g_dwCMD == 0)
        {
            PrintUsage();
            break;
        }
        else if (g_dwCMD & (0x01 << MWCAP_INFO_LIST_CHANNEL_INFO))
        {
            PrintAllChannelInfo();
            break;
        }
        else if (g_dwCMD & (0x01 << MWCAP_INFO_QUERY))
        {
            printf("total %d\n", g_nDevicesNum);

            if (g_nDevicesNum > 0)
                printf("%-15s %-15s %-15s %-15s %-15s %-30s\n",
                       "device path", "firmware ver", "hardware ver", "driver ver", "alsa name","capture device Name");

            int i;
            for (i = 0; i < g_nDevicesNum; i++)
            {
                mr = PrintChannelInfo(g_arrDevicesPath[i]);
                if (mr != MW_SUCCEEDED)
                    printf("print channel info error\n");
            }
        }
        else if(g_dwCMD & (0x01 << MWCAP_INFO_ALL))
        {
            int i;
            for(i = 0; i < g_nDevicesNum; i++)
            {
                if (i != 0)
                    printf("======================================================================\n");
                PrintDeviceInfo(g_arrDevicesPath[i]);
                PrintInputCommon(g_arrDevicesPath[i]);
                PrintInputVideo(g_arrDevicesPath[i]);
                PrintInputAudio(g_arrDevicesPath[i]);
                PrintInputSpecific(g_arrDevicesPath[i]);
                PrintHdmiInfoframe(g_arrDevicesPath[i]);
            }
        }
        else if(g_dwCMD & (0x01 << MWCAP_INFO_DEVICE))
        {
            int i;
            for(i = 0; i < g_nDevicesNum; i++)
            {
                if (i != 0)
                    printf("======================================================================\n");
                PrintDeviceInfo(g_arrDevicesPath[i]);
            }
        }
        else if(g_dwCMD & (0x01 << MWCAP_INFO_INPUT_ALL))
        {
            int i;
            for(i = 0; i < g_nDevicesNum; i++)
            {
                if (i != 0)
                    printf("======================================================================\n");
                PrintInputCommon(g_arrDevicesPath[i]);
                PrintInputVideo(g_arrDevicesPath[i]);
                PrintInputAudio(g_arrDevicesPath[i]);
                PrintInputSpecific(g_arrDevicesPath[i]);
            }
        }
        else if (g_dwCMD & (0x01 << MWCAP_INFO_INPUT_VIDEO))
        {
            int i;
            for(i = 0; i < g_nDevicesNum; i++)
            {
                if (i != 0)
                    printf("======================================================================\n");
                PrintInputCommon(g_arrDevicesPath[i]);
                PrintInputVideo(g_arrDevicesPath[i]);
            }

        }
        else if(g_dwCMD & (0x01 << MWCAP_INFO_INPUT_AUDIO))
        {
            int i;
            for(i = 0; i < g_nDevicesNum; i++)
            {
                if (i != 0)
                    printf("======================================================================\n");
                PrintInputCommon(g_arrDevicesPath[i]);
                PrintInputAudio(g_arrDevicesPath[i]);
            }
        }
        else if(g_dwCMD & (0x01 << MWCAP_INFO_INPUT_SPECIFIC))
        {
            int i;
            for(i = 0; i < g_nDevicesNum; i++)
            {
                if (i != 0)
                    printf("======================================================================\n");
                PrintInputCommon(g_arrDevicesPath[i]);
                PrintInputSpecific(g_arrDevicesPath[i]);
            }
        }
        else if(g_dwCMD & (0x01 << MWCAP_INFO_HDMI))
        {
            int i;
            for(i = 0; i < g_nDevicesNum; i++)
            {
                if (i != 0)
                    printf("======================================================================\n");
                PrintHdmiInfoframe(g_arrDevicesPath[i]);
            }
        }
        else if (g_dwCMD & (0x01 << MWCAP_INFO_V4L2_NAME))
        {
            int i ;
            for (i = 0; i < g_nDevicesNum; i++)
            {
                if (i != 0)
                    printf(" ");

                printf("%s", g_arrDevicesPath[i]);
            }

            printf("\n");
        }
        else if(g_dwCMD & (0x01 << MWCAP_INFO_ALSA_NAME))
        {
            int i;
            for(i = 0; i < g_nDevicesNum; i++)
            {
                int nIndex = GetDeviceIndexByPath(g_arrDevicesPath[i]);
                char szHwName[MAX_HW_NAME_LEN];
                GetAudioHwNameByindex(nIndex, szHwName);

                if (i != 0)
                    printf(" ");

                printf("%s", szHwName);
            }
            printf("\n");
        }
        else if(g_dwCMD & (0x01 << MWCAP_INFO_ALSA_NAME_MPLAYER))
        {
            int i;
            for(i = 0; i < g_nDevicesNum; i++)
            {
                int nIndex = GetDeviceIndexByPath(g_arrDevicesPath[i]);
                char szHwName[MAX_HW_NAME_LEN];
                GetAudioHwNameByindex(nIndex, szHwName);

                char *szTemp;

                szTemp = strchr(szHwName, ':');
                if (szTemp != NULL)
                    *szTemp = '.';

                if (i != 0)
                    printf(" ");

                printf("%s", szHwName);
            }
            printf("\n");
        }
        else if(g_dwCMD & (0x01 << MWCAP_INFO_ALSA_NAME_PA))
        {
            int i;
            for(i = 0; i < g_nDevicesNum; i++)
            {
                int nIndex = GetDeviceIndexByPath(g_arrDevicesPath[i]);
                char szPAName[MAX_PA_NAME_LEN];
                GetAudioPaNameByindex(nIndex, szPAName);

                if (i != 0)
                    printf(" ");

                printf("%s", szPAName);
            }
            printf("\n");
        }
        else if (g_dwCMD & (0x01 << MWCAP_INFO_DEVICE_NUMS))
        {
            printf("%d\n", g_nDevicesNum);
        }


    }while(false);


    MWCaptureExitInstance();

    return 0;
}






