////////////////////////////////////////////////////////////////////////////////
// CONFIDENTIAL and PROPRIETARY software of Magewell Electronics Co., Ltd.
// Copyright (c) 2011-2014 Magewell Electronics Co., Ltd. (Nanjing)
// All rights reserved.
// This copyright notice MUST be reproduced on all authorized copies.
////////////////////////////////////////////////////////////////////////////////
#include "v4l2capturedevice.h"

V4l2CaptureDevice::V4l2CaptureDevice(VideoRenderBase *videoRender) :
    m_nDeviceID(-1),
    m_nSupportFourccCount(0),
    m_threadState(kThreadInvalid),
    m_llFrameCountTotal(0),
    //m_llFrameCountDrop(0),
    m_FrameBuftype(0),
    m_nFrameBufCount(0),
    m_VideoRender(videoRender)
{
    memset(m_aFrameBufs, 0, sizeof(m_aFrameBufs));
}

V4l2CaptureDevice::~V4l2CaptureDevice()
{
}

bool V4l2CaptureDevice::Init(const char *devName)
{
    if (m_nDeviceID != -1)
        Deinit();

    m_nDeviceID = open(devName, O_RDWR);
    if (m_nDeviceID == -1)
        return false;

    strncpy(m_devName, devName, DEVICE_NAME_MAX_LEN);

    GetFrameFormat(m_currentFrameFormat);
    DetectSupportedFourcc();

    return true;
}

void V4l2CaptureDevice::Deinit()
{
    if (m_nDeviceID != -1) {
        close(m_nDeviceID);
        m_nDeviceID = -1;
    }
}

uint32_t *V4l2CaptureDevice::GetSupportedFourcc(int &nCount)
{
    uint32_t *newData = NULL;
    nCount = 0;

    if (m_arSupportFourcc.size() <= 0)
        return NULL;

    newData = new uint32_t[m_arSupportFourcc.size()];
    if (newData == NULL)
        return NULL;

    nCount = m_arSupportFourcc.size();
    for (int i = 0; i < nCount; i++) {
        newData[i] = m_arSupportFourcc[i].mwFourcc;
    }
    return newData;
}

bool V4l2CaptureDevice::SetFrameFormat(MWFrameFormat &format)
{
    if (mwfourcc_to_v4l2(format.fourcc) == 0)
        return false;

    //if (!CheckVideoFormat(pInfo)) TODO
    //  return FALSE;
    v4l2_format vformat;
    memset(&vformat, 0, sizeof(vformat));

    vformat.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    vformat.fmt.pix.width = format.width;
    vformat.fmt.pix.height = format.height;
    vformat.fmt.pix.pixelformat = mwfourcc_to_v4l2(format.fourcc);
    vformat.fmt.pix.field = V4L2_FIELD_ANY;

    if (ioctl(m_nDeviceID, VIDIOC_S_FMT, &vformat) != 0)
        return false;

    //printf("VIDIOC_S_FMT success\n");

    v4l2_streamparm parm;
    memset(&parm, 0, sizeof(parm));

    parm.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    parm.parm.capture.timeperframe.denominator = format.denominator;
    parm.parm.capture.timeperframe.numerator = format.numerator;

    if (ioctl(m_nDeviceID, VIDIOC_S_PARM, &parm) != 0)
        return false;

    //printf("VIDIOC_S_PARM success\n");
    m_currentFrameFormat = format;

    return true;
}

bool V4l2CaptureDevice::GetFrameFormat(MWFrameFormat &format)
{
    v4l2_format vformat;
    memset(&vformat, 0, sizeof(vformat));

    vformat.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

    if (ioctl(m_nDeviceID, VIDIOC_G_FMT, &vformat) != 0)
        return false;

    format.width = vformat.fmt.pix.width;
    format.height = vformat.fmt.pix.height;
    format.fourcc = mwfourcc_from_v4l2(vformat.fmt.pix.pixelformat);

    if (format.fourcc == 0)
        return false;

    v4l2_streamparm parm;
    memset(&parm, 0, sizeof(parm));

    parm.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;

    if (ioctl(m_nDeviceID, VIDIOC_G_PARM, &parm) != 0)
        return false;

    format.denominator = parm.parm.capture.timeperframe.denominator;
    format.numerator = parm.parm.capture.timeperframe.numerator;

    return true;
}

int V4l2CaptureDevice::CreateBuffers(int nCount, v4l2_memory type)
{
    if (m_nFrameBufCount != 0) {
        mw_debug(MW_DEBUG_ERR, "Image buffers is already created!\n");
        return -1;
    }

    if (type == V4L2_MEMORY_MMAP) {
        return CreateBuffersWithMmap(nCount);
    } else if (type == V4L2_MEMORY_USERPTR) {
        return CreateBuffersWithUserptr(nCount);
    }

    return -1;
}

void V4l2CaptureDevice::DestroyBuffers()
{
    if (m_nFrameBufCount == 0)
        return;

    for (int i = 0; i < m_nFrameBufCount; i++) {
        if (m_aFrameBufs[i].addr != NULL && m_aFrameBufs[i].len != 0) {
            munmap(m_aFrameBufs[i].addr, m_aFrameBufs[i].len);
            m_aFrameBufs[i].addr = NULL;
            m_aFrameBufs[i].len = 0;
        }
        if (m_aFrameBufs[i].renderBuf != NULL) {
            m_VideoRender->FreeImageBuffer((VideoRenderImageBuf *)m_aFrameBufs[i].renderBuf);
            m_aFrameBufs[i].renderBuf = NULL;
        }
    }
    m_nFrameBufCount = 0;
}

bool V4l2CaptureDevice::StartStreaming()
{
    if (m_FrameBuftype != V4L2_MEMORY_MMAP && m_FrameBuftype != V4L2_MEMORY_USERPTR)
        return false;

    while (m_threadState == kThreadStarting || m_threadState == kThreadStoping)
        pthread_yield();

    if (m_threadState == kThreadRunning) {
        mw_debug(MW_DEBUG_ERR, "Capture thread is already running!\n");
        return false;
    }

    m_threadState = kThreadStarting;

    if (pthread_create(&m_hThread, NULL, CaptureThreadWrap, this) != 0) {
        mw_debug(MW_DEBUG_ERR, "thread init error: %s\n", strerror(errno));
        return false;
    }

    return true;
}

void V4l2CaptureDevice::StopStreaming()
{
    if (m_threadState != kThreadRunning) {
        mw_debug(MW_DEBUG_ERR, "Capture thread is not running!\n");
        return;
    }

    m_threadState = kThreadStoping;

    while (m_threadState != kThreadInvalid)
        pthread_yield();
}


/* private */
void V4l2CaptureDevice::DetectSupportedFourcc()
{
    struct v4l2_fmtdesc format;

    m_arSupportFourcc.clear();
    for (int i = 0; ; i++) {
        format.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        format.index = i;

        if (ioctl(m_nDeviceID, VIDIOC_ENUM_FMT, &format) < 0)
            return;

        uint32_t mw_fourcc = mwfourcc_from_v4l2(format.pixelformat);
        if (mw_fourcc != 0) {
            FourccMap fourccMap = {
                mw_fourcc, format.pixelformat
            };

            if (!m_arSupportFourcc.has(fourccMap, fourcc_is_equal)) {
                m_arSupportFourcc.append(fourccMap);
            }
        }
    }
}

int V4l2CaptureDevice::CreateBuffersWithMmap(int nCount)
{
    v4l2_requestbuffers vreq;
    int realCount = 0;

    memset(&vreq, 0, sizeof(vreq));

    vreq.count = nCount;
    vreq.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    vreq.memory = V4L2_MEMORY_MMAP;

    if (ioctl(m_nDeviceID, VIDIOC_REQBUFS, &vreq) < 0) {
        mw_debug(MW_DEBUG_ERR, "VIDIOC_REQBUFS error: %s\n", strerror(errno));
        return -1;
    }
    if (vreq.count < 1)
        return -1;

    for (realCount = 0; realCount < (int)vreq.count; ++realCount)  {
        v4l2_buffer vbuf;
        void *addr;

        memset(&vbuf, 0, sizeof(vbuf));

        vbuf.type   = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        vbuf.memory = V4L2_MEMORY_MMAP;
        vbuf.index  = realCount;

        if (ioctl(m_nDeviceID, VIDIOC_QUERYBUF, &vbuf) < 0) {
            mw_debug(MW_DEBUG_ERR, "VIDIOC_QUERYBUF error: %s\n", strerror(errno));
            return -1;
        }

        addr = mmap(NULL, vbuf.length, PROT_READ | PROT_WRITE,
                    MAP_SHARED, m_nDeviceID, vbuf.m.offset);
        if (addr == MAP_FAILED) {
            mw_debug(MW_DEBUG_ERR, "mmap error: %s\n", strerror(errno));
            return -1;
        }

        m_aFrameBufs[realCount].addr = addr;
        m_aFrameBufs[realCount].len = (size_t)vbuf.length;
        m_aFrameBufs[realCount].renderBuf = NULL;
        mw_debug(0, "i=%d v4l2Buf.m.userptr=%lx v4l2Buf.length=%d\n",
                 realCount, (long)addr, vbuf.length);
    }

    m_FrameBuftype = V4L2_MEMORY_MMAP;
    m_nFrameBufCount = realCount;

    return realCount;
}

int V4l2CaptureDevice::CreateBuffersWithUserptr(int nCount)
{
    v4l2_requestbuffers vreq;
    int realCount = 0;

    memset(&vreq, 0, sizeof(vreq));

    vreq.count = nCount;
    vreq.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
    vreq.memory = V4L2_MEMORY_USERPTR;

    if (ioctl(m_nDeviceID, VIDIOC_REQBUFS, &vreq) < 0) {
        mw_debug(MW_DEBUG_ERR, "VIDIOC_REQBUFS error: %s\n", strerror(errno));
        return -1;
    }
    if (vreq.count < 1)
        return -1;

    for (realCount = 0; realCount < (int)vreq.count; realCount++) {
        VideoRenderImageBuf *renderBuf =
                m_VideoRender->AllocImageBuffer(VIDEO_RENDER_IMAGEBUF_VIRTUAL);

        if (renderBuf == NULL)
            return -1;

        m_aFrameBufs[realCount].addr = NULL;
        m_aFrameBufs[realCount].renderBuf = renderBuf;
        m_aFrameBufs[realCount].len = renderBuf->size;
    }

    m_FrameBuftype = V4L2_MEMORY_USERPTR;
    m_nFrameBufCount = realCount;

    return realCount;
}

void *V4l2CaptureDevice::CaptureThreadWrap(void *data)
{
    V4l2CaptureDevice *device = reinterpret_cast<V4l2CaptureDevice *>(data);
    int ret;

    do {
        ret = 0;
        struct v4l2_buffer v4l2Buf;

        for (int i = 0; i < device->m_nFrameBufCount; i++) {
            memset(&v4l2Buf, 0, sizeof(v4l2Buf));

            v4l2Buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            v4l2Buf.memory = (v4l2_memory)device->m_FrameBuftype;
            v4l2Buf.index = i;
            if (device->m_FrameBuftype == V4L2_MEMORY_USERPTR) {
                v4l2Buf.m.userptr =
                        (unsigned long)(((VideoRenderImageBuf *)device->m_aFrameBufs[i].renderBuf)->buf);
                v4l2Buf.length = device->m_aFrameBufs[i].len;
                mw_debug(0, "i=%d v4l2Buf.m.userptr=%lx v4l2Buf.length=%d\n",
                         i, v4l2Buf.m.userptr, v4l2Buf.length);
            }
            if (-1 == ioctl(device->m_nDeviceID, VIDIOC_QBUF, &v4l2Buf)) {
                mw_debug(MW_DEBUG_ERR, "VIDIOC_QBUF error: %s\n", strerror(errno));
                ret = -1;
                break;
            }
        }
        if (ret < 0) {
            ret = -1;
            break;
        }

        v4l2_buf_type vtype;
        vtype = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        if (-1 == ioctl(device->m_nDeviceID, VIDIOC_STREAMON, &vtype)) {
            mw_debug(MW_DEBUG_ERR, "VIDIOC_STREAMON error: %s\n", strerror(errno));
            ret = -1;
            break;
        }

        device->m_threadState = kThreadRunning;

        ret = device->CaptureThreadProc();

        vtype = V4L2_BUF_TYPE_VIDEO_CAPTURE;
        if (-1 == ioctl(device->m_nDeviceID, VIDIOC_STREAMOFF, &vtype)) {
            mw_debug(MW_DEBUG_ERR, "VIDIOC_STREAMOFF error: %s\n", strerror(errno));
            ret = -1;
            break;
        }
    } while (0);

    device->m_threadState = kThreadInvalid;

    mw_debug(MW_DEBUG_INFO, "Capture thread exit\n");
    return (void *)(long)ret;
}

int V4l2CaptureDevice::CaptureThreadProc()
{
    struct v4l2_buffer v4l2Buf;
    struct timeval tv_init;
    int skip_count = 4;

    tv_init.tv_sec = (m_currentFrameFormat.numerator * 2) / m_currentFrameFormat.denominator;
    tv_init.tv_usec = (((m_currentFrameFormat.numerator * 2) % m_currentFrameFormat.denominator) * 1000000LL)
        / m_currentFrameFormat.denominator;

    m_llFrameCountTotal = 0;
    gettimeofday(&m_captureStartTime, NULL);

    while (m_threadState != kThreadStoping) {
        for (;;) {
            fd_set fds;
            struct timeval tv;
            int ret;

            FD_ZERO(&fds);
            FD_SET(m_nDeviceID, &fds);

            /* Timeout. */
            tv = tv_init;

            ret = select(m_nDeviceID + 1, &fds, NULL, NULL, &tv);

            if (-1 == ret) {
                if (EINTR == errno)
                    continue;
                mw_debug(MW_DEBUG_ERR, "select error: %s\n", strerror((errno)));
                return -1;
            }

            if (0 == ret) {
                mw_debug(MW_DEBUG_ERR, "select timeout\n");
                break;
            }

            memset(&v4l2Buf, 0, sizeof(v4l2Buf));
            v4l2Buf.type = V4L2_BUF_TYPE_VIDEO_CAPTURE;
            v4l2Buf.memory = (v4l2_memory)m_FrameBuftype;

            if (-1 == ioctl(m_nDeviceID, VIDIOC_DQBUF, &v4l2Buf)) {
                switch (errno) {
                case EAGAIN:
                    break;

                case EIO:

                default:
                    mw_debug(MW_DEBUG_ERR, "VIDIOC_DQBUF error: %s\n", strerror((errno)));
                    return -errno;
                }
            }

            m_llFrameCountTotal++;

            if (m_llFrameCountTotal == skip_count) {
                gettimeofday(&m_captureStartTime, NULL);
            }

            if (/*m_llFrameCountTotal % 10 == 0 && m_llFrameCountTotal != 0*/
                    m_llFrameCountTotal > skip_count + 10) {
                struct timeval tv_cur;
                double timetotal;
                gettimeofday(&tv_cur, NULL);

                if (tv_cur.tv_usec < m_captureStartTime.tv_usec) {
                    timetotal = (double)(tv_cur.tv_sec - m_captureStartTime.tv_sec - 1) +
                            (double)(tv_cur.tv_usec + 1000000 - m_captureStartTime.tv_usec) / 1000000;
                } else {
                    timetotal = (double)(tv_cur.tv_sec - m_captureStartTime.tv_sec) +
                            (double)(tv_cur.tv_usec - m_captureStartTime.tv_usec) / 1000000;
                }

                m_VideoRender->DisplayFps(m_llFrameCountTotal - skip_count,
                                          (double)(m_llFrameCountTotal - skip_count) / timetotal);
            }

            if (m_aFrameBufs[v4l2Buf.index].addr != NULL && m_aFrameBufs[v4l2Buf.index].len != 0) {
                m_VideoRender->DisplayFrame(m_aFrameBufs[v4l2Buf.index].addr,
                        m_aFrameBufs[v4l2Buf.index].len);
            } else if (m_aFrameBufs[v4l2Buf.index].renderBuf != NULL) {
                m_VideoRender->DisplayFrame((VideoRenderImageBuf *)m_aFrameBufs[v4l2Buf.index].renderBuf);
            }

            if (-1 == ioctl(m_nDeviceID, VIDIOC_QBUF, &v4l2Buf)) {
                mw_debug(MW_DEBUG_ERR, "VIDIOC_QBUF error: %s\n", strerror(errno));
                return -1;
            }

            break;
        }
    }
    return 0;
}
