////////////////////////////////////////////////////////////////////////////////
// CONFIDENTIAL and PROPRIETARY software of Magewell Electronics Co., Ltd.
// Copyright (c) 2011-2014 Magewell Electronics Co., Ltd. (Nanjing)
// All rights reserved.
// This copyright notice MUST be reproduced on all authorized copies.
////////////////////////////////////////////////////////////////////////////////
#ifndef V4L2CAPTUREDEVICE_H
#define V4L2CAPTUREDEVICE_H

#ifdef __cplusplus
extern "C" {
#endif
#include <string.h>
#include <stdlib.h>
#include <stdio.h>

#include <errno.h>
#include <fcntl.h>
#include <unistd.h>

#include <sys/ioctl.h>
#include <sys/mman.h>

#include <linux/videodev2.h>
#ifdef __cplusplus
}
#endif

#include "videorenderbase.h"

typedef struct _MWFrameFormat {
    int         width;
    int         height;
    uint32_t    fourcc;
    int         denominator;
    int         numerator;
} MWFrameFormat;

class V4l2CaptureDevice
{
#define DEVICE_NAME_MAX_LEN     (32)
#define V4l2_MAX_BUFFERS        (4)
public:
    V4l2CaptureDevice(VideoRenderBase *videoRender);
    virtual ~V4l2CaptureDevice();

    bool Init(const char *devName);
    void Deinit();

    uint32_t *GetSupportedFourcc(int &nCount);

    bool SetFrameFormat(MWFrameFormat &format);
    bool GetFrameFormat(MWFrameFormat &format);

    int CreateBuffers(int nCount, v4l2_memory type);
    void DestroyBuffers();

    bool StartStreaming();
    void StopStreaming();

    static uint32_t mwfourcc_from_v4l2(uint32_t v4l2_fourcc)
    {
        switch (v4l2_fourcc) {
        case V4L2_PIX_FMT_RGB24:
            return MWFOURCC_RGB24;
        case V4L2_PIX_FMT_RGB32:
            return MWFOURCC_RGBA;
        case V4L2_PIX_FMT_BGR24:
            return MWFOURCC_BGR24;
        case V4L2_PIX_FMT_BGR32:
            return MWFOURCC_BGRA;
        case V4L2_PIX_FMT_YUYV:
            return MWFOURCC_YUY2;
        case V4L2_PIX_FMT_UYVY:
            return MWFOURCC_UYVY;
        case V4L2_PIX_FMT_YUV420:
            return MWFOURCC_I420;
        case V4L2_PIX_FMT_YVU420:
            return MWFOURCC_YV12;
        case V4L2_PIX_FMT_NV12:
            return MWFOURCC_NV12;
        }

        return 0;
    }

    static uint32_t mwfourcc_to_v4l2(uint32_t mwfourcc)
    {
        switch (mwfourcc) {
        case MWFOURCC_RGB24:
            return V4L2_PIX_FMT_RGB24;
        case MWFOURCC_RGBA:
            return V4L2_PIX_FMT_RGB32;
        case MWFOURCC_BGR24:
            return V4L2_PIX_FMT_BGR24;
        case MWFOURCC_BGRA:
            return V4L2_PIX_FMT_BGR32;
        case MWFOURCC_YUYV:
        case MWFOURCC_YUY2:
            return V4L2_PIX_FMT_YUYV;
        case MWFOURCC_UYVY:
            return V4L2_PIX_FMT_UYVY;
        case MWFOURCC_I420:
            return V4L2_PIX_FMT_YUV420;
        case MWFOURCC_YV12:
            return V4L2_PIX_FMT_YVU420;
        case MWFOURCC_NV12:
            return V4L2_PIX_FMT_NV12;
        }

        return 0;
    }

private:
    void DetectSupportedFourcc();

    int CreateBuffersWithMmap(int nCount);
    int CreateBuffersWithUserptr(int nCount);


    static void *CaptureThreadWrap(void *data);
    int CaptureThreadProc();

private:
    typedef struct _V4l2CaptureBuf {
        void                    *addr;
        size_t                  len;

        void                    *renderBuf;
    } V4l2CaptureBuf;

    int                         m_nDeviceID;
    char                        m_devName[DEVICE_NAME_MAX_LEN];

    typedef struct _FourccMap {
        uint32_t            mwFourcc;
        uint32_t            v4l2Froucc;
    } FourccMap;

    static bool fourcc_is_equal(FourccMap &item1, FourccMap &item2) {
        return item1.mwFourcc == item2.mwFourcc;
    }

    MWArray<FourccMap, 16>      m_arSupportFourcc;
    int                         m_nSupportFourccCount;

    MWFrameFormat               m_currentFrameFormat;

    /* capture process */
    pthread_t                   m_hThread;
    enum { kThreadInvalid, kThreadRunning, kThreadStarting, kThreadStoping};
    int                         m_threadState;
    struct timeval              m_captureStartTime;
    long long                   m_llFrameCountTotal;
    //long long                   m_llFrameCountDrop;


    /* frame buffers */
    int                         m_FrameBuftype; /* v4l2_memory */
    int                         m_nFrameBufCount;
    V4l2CaptureBuf              m_aFrameBufs[V4l2_MAX_BUFFERS];

    VideoRenderBase             *m_VideoRender;
};

#endif // V4L2CAPTUREDEVICE_H
